// dz_utils: extract 1.0a-2007-01-03
//
// note: this file is PUBLIC DOMAIN
// feel free to do whatever you want with it

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

// defs.h

typedef long INT32;
typedef short INT16;
typedef char INT8;
typedef unsigned long UINT32;
typedef unsigned short UINT16;
typedef unsigned char UINT8;

#define STR_LENGTH 1024

#define FIND_ANY 0
#define FIND_TYPE 0x0001
#define FIND_SUBTYPE 0x0002
#define FIND_GROUP 0x0004
#define FIND_INSTANCE 0x0008

typedef struct tagFindRec {
	UINT32 type;
	UINT32 subtype;
	UINT32 group;
	UINT32 instance;
	UINT32 size;
	UINT32 index;
} FindRec, *pFindRec;

// extract.h

void loadtoc(INT8 *s);
UINT8 *unPack(UINT8 *b, UINT32 s1, UINT32 s2);
UINT8 *get2(UINT32 o, UINT32 *s);
UINT32 find2(UINT32 cmd, pFindRec pFind);
void do_extract(int gb, UINT32 g, int tb, UINT32 t, int ib, UINT32 i, int sb, UINT32 s, int op);
UINT32 hex2int(INT8 *s);
int opt_parse(INT8 *o, int x);

// global vars

UINT32 gSize, gVer, *gToc;
INT8 gName[STR_LENGTH];

// dowah dowah

void loadtoc(INT8 *s)
{
	int status = 0;
	UINT32 xStart, xSize, xVer, *pToc;
	FILE *f = fopen(s, "rb");

	gToc = NULL;
	if (f != NULL) {
		fseek(f, 0x28, 0);
		fread(&xStart, 4, 1, f);			// !le!
		fread(&xSize, 4, 1, f);				// !le!
		fseek(f, 0x3C, 0);
		fread(&xVer, 4, 1, f);				// subversion (1/2) !le!

		if (xSize != 0) {
			fseek(f, xStart, 0);
			pToc = (UINT32 *) malloc(xSize);
			if (pToc != NULL) {
				fread(pToc, xSize, 1, f);	// load toc
				status = 1;
			}
		}

		if (status) {
			gToc = pToc;
			gSize = xSize;
			gVer = xVer;
			strcpy(gName, s);
		}

		fclose(f);
	}
}

UINT8 *unPack(UINT8 *b, UINT32 s1, UINT32 s2)
{
	UINT32 x1, x2, i, o, q;
	UINT8 a, r, d;
	UINT8 *src, *dst;

	src = b;
	dst = (UINT8 *) malloc(s2);					// alloc destination buffer
	if (dst != NULL) {
		x1 = 9;		// src index
		x2 = 0;		// dest index

		while (x1 < s1) {

			// decode ctrl bytes
			a = src[x1++];						// ctrl

			if ((a & 0x80) == 0) {				// 0ooqqqrr oooooooo
				r = a & 3;
				q = ((a >> 2) & 7) + 3;
				o = ((a << 3) & 0x300) + src[x1++];

			} else if ((a & 0x40) == 0) {   	// 10qqqqqq rroooooo oooooooo
				q = (a & 0x3F) + 4;
				a = src[x1++];
				r = (a >> 6) & 3;
				o = ((a << 8) & 0x3F00) + src[x1++];

			} else if ((a & 0x20) == 0) {		// 110oqqrr oooooooo oooooooo qqqqqqqq
				r = a & 3;
				o = ((a << 12) & 0x10000) + (src[x1++] << 8);
				o += src[x1++];
				q = ((a << 6) & 0x300) + src[x1++] + 5;

			} else {							// 111rrrrr / 111111rr
				r = ((a & 0x1F) + 1) << 2;
				if ((a & 0x1C) == 0x1C) r = a & 3;
				q = 0;
			}

			// read block
			if (x2 + r <= s2) {
				for (i = 1; i <= r; i++) {		// read through a literal block
					dst[x2++] = src[x1++];
				}
			} else {
				x1 += r;
				x2 += r;
			}

			// repeat block
			o = (x2 - o) - 1;					// repeat index = current index (of last read byte) - offset
			if (x2 + q <= s2) {
				for (i = 1; i <= q; i++) {
					dst[x2++] = dst[o++];
				}
			} else
				x2 += q;
		}

		// kaboom
		assert(x1 == s1);
		assert(x2 == s2);
	}
	return dst;
}

UINT8 *get2(UINT32 o, UINT32 *s)
{
	FILE *f;
	UINT32 s1, s2;
	UINT8 *buf, *b2, cmp = 0;

	buf = NULL;
	s1 = *s;
	if ((f = fopen(gName, "rb")) != NULL) {
		buf = (UINT8 *) malloc(s1);
		if (buf != NULL) {
			fseek(f, o, 0);
			fread(buf, s1, 1, f);			// get file data
			s2 = s1;
			b2 = buf;						// b2 = buf

			// get compression header info (if any)
			if (s1 >= 10) {					// header + 1 ctrl
				s2 = *(UINT32*) (&buf[0]);
				if (s2 == s1) {				// confirm size
					if (b2[4] == 0x10 && b2[5] == 0xFB) {			// check cookie
						s2 = (b2[6] << 16) + (b2[7] << 8) + b2[8];	// real size
						cmp = 1;
					}
				}
			}

			// get the data
			if (cmp) {
				b2 = unPack(buf, s1, s2);
				if (b2 != NULL) {
					free(buf);		// discard old buf
					buf = b2;		// buf = new buf
					*s = s2;		// size = new size
				}
			}
		}
		fclose(f);
	}
	return buf;
}

UINT32 find2(UINT32 cmd, pFindRec pFind)
{
	UINT32 s, v, x, *t;
	UINT32 eType, eSubType, eGroup, eInstance, eOffset, eSize;

	s = (gSize >> 2);
	v = gVer;
	t = gToc;
	x = pFind->index;
	eSubType = 0;

	// find, or not
	while (x < s) {
		eType = t[x++];
		eGroup = t[x++];
		eInstance = t[x++];
		if (v > 1) eSubType = t[x++];
		eOffset = t[x++];
		eSize = t[x++];

		// todo: sort out this mess
		if (!( ((cmd & FIND_TYPE) && (eType != pFind->type)) ||
			((cmd & FIND_SUBTYPE) && (eSubType != pFind->subtype)) ||
			((cmd & FIND_GROUP) && (eGroup != pFind->group)) ||
			((cmd & FIND_INSTANCE) && (eInstance != pFind->instance))
		) ) {
			pFind->type = eType;
			pFind->subtype = eSubType;
			pFind->group = eGroup;
			pFind->instance = eInstance;
			pFind->size = eSize;
			pFind->index = x;
			return(eOffset);
		}
	}
	return(0);
}

void do_extract(int gb, UINT32 g, int tb, UINT32 t, int ib, UINT32 i, int sb, UINT32 s, int op)
{
	UINT32 o, ft;
	FindRec fr = { 0, 0, 0, 0, 0, 0 };
	UINT8 *buf;
	INT8 n[STR_LENGTH];
	FILE *f;

	ft = FIND_ANY;					// default = find anything
	if (gb != 0) {
		ft += FIND_GROUP;
		fr.group = g;
	}
	if (tb != 0) {
		ft += FIND_TYPE;
		fr.type = t;
	}
	if (ib != 0) {
		ft += FIND_INSTANCE;
		fr.instance = i;
	}
	if (sb != 0) {
		ft += FIND_SUBTYPE;
		fr.subtype = s;
	}

	o = find2(ft, &fr);				// find first
	while (o != 0) {
		// get data
		buf = get2(o, &fr.size);
		if (buf != NULL) {

			// make a filename
			sprintf(n, "%08X-%08X-%08X-%08X.dat", fr.group, fr.type, fr.instance, fr.subtype);
			if ((op & 1) && (fr.subtype == 0))
				sprintf(n, "%08X-%08X-%08X.dat", fr.group, fr.type, fr.instance);

			// dump buf -> file
			f = fopen(n, "wb");
			if (f != NULL) {
				fwrite(buf, 1, fr.size, f);
				fclose(f);
			}

			// discard buf
			free(buf);
		}
		o = find2(ft, &fr);			// find next
	}
}

UINT32 hex2int(INT8 *s)
{
	INT8 *s2 = s;
	UINT32 v;

	if (strlen(s) > 2) {
		// hexfix?
		if (stricmp(s, "0x") == 0) {
			// ignore
			s2 = s + 2;
		}
	}
	// assume this is hex
	if (sscanf(s2, "%x", &v) == 1)
		return v;
	else
		return 0;
}

int opt_parse(INT8 *o, int x)
{
	if (strcmp(o, "-g") == 0) return 1;
	if (strcmp(o, "-t") == 0) return 2;
	if (strcmp(o, "-i") == 0) return 3;
	if (strcmp(o, "-s") == 0) return 4;
	if (strcmp(o, "-q") == 0) return 5;
	if (x == 1) return 101;
	if (x == 2) return 102;
	if (x == 3) return 103;
	if (x == 4) return 104;
	return 0;
}

int main(int argc, char *argv[])
{
	FILE *outFile = stdout;
	int it, x = 0, op = 0;
	int gb = 0, tb = 0, ib = 0, sb = 0;
	UINT32 g, t, i, s;
	INT8 *sf = NULL;

	if (argc > 1) {
		for (it = 1; it < argc; it++) {
			x = opt_parse(argv[it], x);
			if (x == 0) {
				sf = argv[it];
			}
			if (x == 101) {
				g = hex2int(argv[it]);
				gb = 1;
			}
			if (x == 102) {
				t = hex2int(argv[it]);
				tb = 1;
			}
			if (x == 103) {
				i = hex2int(argv[it]);
				ib = 1;
			}
			if (x == 104) {
				s = hex2int(argv[it]);
				sb = 1;
			}
			if (x == 5) op += 1;
		}
		loadtoc(sf);
		if (gToc != NULL) {
			do_extract(gb, g, tb, t, ib, i, sb, s, op);
			free(gToc);
		}
	} else {
		printf("extract usage:\n\n");
		printf("extract [-g group] [-t type] [-i instance] [-s subtype] [-q] package\n");
		printf("  -g <hexvalue>     extract matches of specified group\n");
		printf("  -t <hexvalue>     extract matches of specified type\n");
		printf("  -i <hexvalue>     extract matches of specified instance\n");
		printf("  -s <hexvalue>     extract matches of specified subtype\n");
		printf("  -q                suppress subtype in filename (if 0)\n\n");
		printf("for example:\n");
		printf("  extract -g 0x7FD46CD0 -t 0x42484156 objects.package\n");
		printf("  extract -g 7FD46CD0 -t 42434F4E -i 106 objects.package\n\n");
	}
	if (outFile != stdout) fclose(outFile);
	return 0;
}
