// dz_utils: ls 1.0b-2007-01-03
//
// note: this file is PUBLIC DOMAIN
// feel free to do whatever you want with it

// little endian gotchas marked: !le!
// todo: fix gotchas

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

// defs.h

typedef long INT32;
typedef short INT16;
typedef char INT8;
typedef unsigned long UINT32;
typedef unsigned short UINT16;
typedef unsigned char UINT8;

#define STR_LENGTH 1024

#define FIND_ANY 0
#define FIND_TYPE 0x0001
#define FIND_SUBTYPE 0x0002
#define FIND_GROUP 0x0004
#define FIND_INSTANCE 0x0008

typedef struct tagFindRec {
	UINT32 type;
	UINT32 subtype;
	UINT32 group;
	UINT32 instance;
	UINT32 size;
	UINT32 index;
} FindRec, *pFindRec;

typedef struct tagFindRecLinked {
	FindRec fr;
	INT8 name[STR_LENGTH];
	struct tagFindRecLinked* prev;
	struct tagFindRecLinked* next;
} FindRecLinked, *pFindRecLinked;

// ls.h

void loadtoc(INT8 *s);
UINT8 *unPack(UINT8 *b, UINT32 s1, UINT32 s2);
UINT8 *get2(UINT32 o, UINT32 *s);
UINT32 find2(UINT32 cmd, pFindRec pFind);
UINT8 *get_name(UINT32 x, UINT32 *s, UINT32 t);
void lower(INT8 *s, INT8 *s2);
void do_list(FILE *f, int gb, UINT32 g, int tb, UINT32 t, INT8 *sp, int or, int rv);
UINT32 hex2int(INT8 *s);
int opt_parse(INT8 *o, int x);

// global vars

UINT32 gSize, gVer, *gToc;
INT8 gName[STR_LENGTH];
UINT8 gStr[STR_LENGTH];

// dowah dowah

void loadtoc(INT8 *s)
{
	int status = 0;
	UINT32 xStart, xSize, xVer, *pToc;
	FILE *f = fopen(s, "rb");

	gToc = NULL;
	if (f != NULL) {
		fseek(f, 0x28, 0);
		fread(&xStart, 4, 1, f);			// !le!
		fread(&xSize, 4, 1, f);				// !le!
		fseek(f, 0x3C, 0);
		fread(&xVer, 4, 1, f);				// subversion (1/2) !le!

		if (xSize != 0) {
			fseek(f, xStart, 0);
			pToc = (UINT32 *) malloc(xSize);
			if (pToc != NULL) {
				fread(pToc, xSize, 1, f);	// load toc
				status = 1;
			}
		}

		if (status) {
			gToc = pToc;
			gSize = xSize;
			gVer = xVer;
			strcpy(gName, s);
		}

		fclose(f);
	}
}

UINT8 *unPack(UINT8 *b, UINT32 s1, UINT32 s2)
{
	UINT32 x1, x2, i, o, q;
	UINT8 a, r, d;
	UINT8 *src, *dst;

	src = b;
	dst = (UINT8 *) malloc(s2);					// alloc destination buffer
	if (dst != NULL) {
		x1 = 9;		// src index
		x2 = 0;		// dest index

		while (x1 < s1) {

			// decode ctrl bytes
			a = src[x1++];						// ctrl

			if ((a & 0x80) == 0) {				// 0ooqqqrr oooooooo
				r = a & 3;
				q = ((a >> 2) & 7) + 3;
				o = ((a << 3) & 0x300) + src[x1++];

			} else if ((a & 0x40) == 0) {   	// 10qqqqqq rroooooo oooooooo
				q = (a & 0x3F) + 4;
				a = src[x1++];
				r = (a >> 6) & 3;
				o = ((a << 8) & 0x3F00) + src[x1++];

			} else if ((a & 0x20) == 0) {		// 110oqqrr oooooooo oooooooo qqqqqqqq
				r = a & 3;
				o = ((a << 12) & 0x10000) + (src[x1++] << 8);
				o += src[x1++];
				q = ((a << 6) & 0x300) + src[x1++] + 5;

			} else {							// 111rrrrr / 111111rr
				r = ((a & 0x1F) + 1) << 2;
				if ((a & 0x1C) == 0x1C) r = a & 3;
				q = 0;
			}

			// read block
			if (x2 + r <= s2) {
				for (i = 1; i <= r; i++) {		// read through a literal block
					dst[x2++] = src[x1++];
				}
			} else {
				x1 += r;
				x2 += r;
			}

			// repeat block
			o = (x2 - o) - 1;					// repeat index = current index (of last read byte) - offset
			if (x2 + q <= s2) {
				for (i = 1; i <= q; i++) {
					dst[x2++] = dst[o++];
				}
			} else
				x2 += q;
		}

		// kaboom
		assert(x1 == s1);
		assert(x2 == s2);
	}
	return dst;
}

UINT8 *get2(UINT32 o, UINT32 *s)
{
	FILE *f;
	UINT32 s1, s2;
	UINT8 *buf, *b2, cmp = 0;

	buf = NULL;
	s1 = *s;
	if ((f = fopen(gName, "rb")) != NULL) {
		buf = (UINT8 *) malloc(s1);
		if (buf != NULL) {
			fseek(f, o, 0);
			fread(buf, s1, 1, f);			// get file data
			s2 = s1;
			b2 = buf;						// b2 = buf

			// get compression header info (if any)
			if (s1 >= 10) {					// header + 1 ctrl
				s2 = *(UINT32*) (&buf[0]);
				if (s2 == s1) {				// confirm size
					if (b2[4] == 0x10 && b2[5] == 0xFB) {			// check cookie
						s2 = (b2[6] << 16) + (b2[7] << 8) + b2[8];	// real size
						cmp = 1;
					}
				}
			}

			// get the data
			if (cmp) {
				b2 = unPack(buf, s1, s2);
				if (b2 != NULL) {
					free(buf);		// discard old buf
					buf = b2;		// buf = new buf
					*s = s2;		// size = new size
				}
			}
		}
		fclose(f);
	}
	return buf;
}

UINT32 find2(UINT32 cmd, pFindRec pFind)
{
	UINT32 s, v, x, *t;
	UINT32 eType, eSubType, eGroup, eInstance, eOffset, eSize;

	s = (gSize >> 2);
	v = gVer;
	t = gToc;
	x = pFind->index;
	eSubType = 0;

	// find, or not
	while (x < s) {
		eType = t[x++];
		eGroup = t[x++];
		eInstance = t[x++];
		if (v > 1) eSubType = t[x++];
		eOffset = t[x++];
		eSize = t[x++];

		// todo: sort out this mess
		if (!( ((cmd & FIND_TYPE) && (eType != pFind->type)) ||
			((cmd & FIND_SUBTYPE) && (eSubType != pFind->subtype)) ||
			((cmd & FIND_GROUP) && (eGroup != pFind->group)) ||
			((cmd & FIND_INSTANCE) && (eInstance != pFind->instance))
		) ) {
			pFind->type = eType;
			pFind->subtype = eSubType;
			pFind->group = eGroup;
			pFind->instance = eInstance;
			pFind->size = eSize;
			pFind->index = x;
			return(eOffset);
		}
	}
	return(0);
}

UINT8 *get_name(UINT32 x, UINT32 *s, UINT32 t)
{
	// returns name in gStr
	UINT8 *buf;
	UINT32 p;

	buf = get2(x, s);
	if (buf != NULL) {
		if (t == 0x4E524546) {								// NREF
			strncpy((INT8 *) gStr, (INT8 *) &buf[0], *s);
			gStr[*s] = 0;
		} else if (t == 0xE86B1EEF) {						// Compression
			gStr[0] = 0;
		} else if (t == 0x9012468A || t == 0x9012468B) {	// Lua
			p = *(UINT32 *) (&buf[4]);						// !le!
			strncpy((INT8 *) gStr, (INT8 *) &buf[8], p);
			gStr[p] = 0;
		} else {
			strncpy((INT8 *) gStr, (INT8 *) &buf[0], 0x40);
			gStr[0x40] = 0;
		}
		free(buf);
	}
	return gStr;
}

void lower(INT8 *s, INT8 *s2)
{
	// s = source, s2 = destination: destination = lower(source)
	INT8 c;
	int x;

	strcpy(s2, s);
	for (x = 0; x < strlen(s2); x++) {
		c = s2[x];
		if (c >= 0x41 && c <= 0x5A) s2[x] = c + 0x20;
	}
}

void do_list(FILE *f, int gb, UINT32 g, int tb, UINT32 t, INT8 *sp, int or, int rv)
{
	UINT32 o, ft;
	INT8 *s;
	INT8 s1[STR_LENGTH], s2[STR_LENGTH];
	int b;
	FindRec fr = { 0, 0, 0, 0, 0, 0 };
	pFindRecLinked head = NULL, tail = NULL, it, mt, nt, pt;

	ft = FIND_ANY;					// default = find anything
	if (gb != 0) {
		ft += FIND_GROUP;
		fr.group = g;
	}
	if (tb != 0) {
		ft += FIND_TYPE;
		fr.type = t;
	}

	o = find2(ft, &fr);				// find first
	while (o != 0) {
		s = (INT8 *) get_name(o, &fr.size, fr.type);
		b = 1;

		// matches specified name?
		if (sp != NULL) {
			b = 0;
			lower(s, s1);			// ignore case
			lower(sp, s2);
			if (strstr(s1, s2) != NULL) b = 1;
		}

		// add this to the list (insert according to sort)
		if (b != 0) {

			// new frl
			nt = (pFindRecLinked) malloc( sizeof(FindRecLinked) );
			nt->fr.group = fr.group;
			nt->fr.index = o;
			nt->fr.instance = fr.instance;
			nt->fr.size = fr.size;
			nt->fr.subtype = fr.subtype;
			nt->fr.type = fr.type;
			strcpy(nt->name, s);

			// find insert point
			it = head;
			mt = NULL;
			while (it != NULL && mt == NULL) {
				if (or == 1) {		// group
					if (fr.group < it->fr.group) mt = it;
				}
				if (or == 2) {		// type
					if (fr.type < it->fr.type) mt = it;
				}
				if (or == 3) {		// subtype
					if (fr.subtype < it->fr.subtype) mt = it;
				}
				if (or == 4) {		// instance
					if (fr.instance < it->fr.instance) mt = it;
				}
				if (or == 5) {		// index
					if (o < it->fr.index) mt = it;
				}
				if (or == 6) {		// size
					if (fr.size < it->fr.size) mt = it;
				}
				if (or == 7) {		// name
					if (stricmp(s, it->name) < 0) mt = it;
				}
				it = it->next;		// next
			}

			// no match, append
			if (mt == NULL) {
				nt->prev = tail;	// link new frl
				nt->next = NULL;
				if (tail != NULL) tail->next = nt;
				tail = nt;
				if (head == NULL) head = tail;

			// insert
			} else {
				pt = mt->prev;
				nt->prev = pt;		// link new frl
				nt->next = mt;
				if (pt != NULL)
					pt->next = nt;
				else
					head = nt;
				mt->prev = nt;
			}
		}
		o = find2(ft, &fr);			// find next
	}

	// print out list, clean up
	if (head != NULL) {
		// header
		fprintf(f, "Group      Type       Subtype    Instance   Index      Size       Name\n");

		it = head;
		if (rv != 0) it = tail;

		while (it != NULL) {
			pt = it;
			fprintf(f, "0x%08X 0x%08X 0x%08X 0x%08X 0x%08X 0x%08X %s\n", \
				it->fr.group, it->fr.type, it->fr.subtype, it->fr.instance, it->fr.index, \
				it->fr.size, it->name);

			if (rv != 0)		// next
				it = it->prev;
			else
				it = it->next;

			free(pt);			// discard frl
		}
	}
}

UINT32 hex2int(INT8 *s)
{
	INT8 *s2 = s;
	UINT32 v;

	if (strlen(s) > 2) {
		// hexfix?
		if (stricmp(s, "0x") == 0) {
			// ignore
			s2 = s + 2;
		}
	}
	// assume this is hex
	if (sscanf(s2, "%x", &v) == 1)
		return v;
	else
		return 0;
}

int opt_parse(INT8 *o, int x)
{
	if (strcmp(o, "-g") == 0) return 1;
	if (strcmp(o, "-t") == 0) return 2;
	if (strcmp(o, "-n") == 0) return 3;
	if (strcmp(o, "-o") == 0) return 4;
	if (strcmp(o, "-r") == 0) return 5;
	if (x == 1) return 101;
	if (x == 2) return 102;
	if (x == 3) return 103;
	if (x == 4) return 104;
	return 0;
}

int main(int argc, char *argv[])
{
	FILE *outFile = stdout;
	int i, x = 0, pf = 0, or = 0, rv = 0;
	int gb = 0, tb = 0;
	UINT32 g, t;
	INT8 *s = NULL, *sp = NULL;

	if (argc > 1) {
		for (i = 1; i < argc; i++) {
			x = opt_parse(argv[i], x);
			if (x == 0) {
				if (pf == 0) {
					s = argv[i];
					pf = 1;
				} else if (outFile == stdout) {
					outFile = fopen(argv[i], "w");
					if (outFile == NULL) outFile = stdout;
				}
			}
			if (x == 101) {
				g = hex2int(argv[i]);
				gb = 1;
			}
			if (x == 102) {
				t = hex2int(argv[i]);
				tb = 1;
			}
			if (x == 103) sp = argv[i];
			if (x == 104) {
				if (strcmp(argv[i], "g") == 0) or = 1;
				else if (strcmp(argv[i], "t") == 0) or = 2;
				else if (strcmp(argv[i], "u") == 0) or = 3;
				else if (strcmp(argv[i], "i") == 0) or = 4;
				else if (strcmp(argv[i], "x") == 0) or = 5;
				else if (strcmp(argv[i], "s") == 0) or = 6;
				else if (strcmp(argv[i], "n") == 0) or = 7;
			}
			if (x == 5) rv = 1;
		}
		loadtoc(s);
		if (gToc != NULL) {
			do_list(outFile, gb, g, tb, t, sp, or, rv);
			free(gToc);
		}
	} else {
		printf("ls usage:\n\n");
		printf("ls [-g group] [-t type] [-n name] [-o g|t|u|i|x|s|n] [-r] package [output]\n");
		printf("  -g <hexvalue>     list matches of specified group\n");
		printf("  -t <hexvalue>     list matches of specified type\n");
		printf("  -n <string>       list matches of specified name\n");
		printf("  -o <order>        list in order of specified field\n");
		printf("     g = group, t = type, u = subtype, i = instance, x = index,\n");
		printf("     s = size, n = name\n");
		printf("  -r                list in descending order (ascending = default)\n\n");
		printf("for example:\n");
		printf("  ls -t 0x42484156 -o n objects.package dir.txt\n");
		printf("  ls -g 7FD46CD0 -t 42434F4E -o i objects.package dir.txt\n\n");
	}
	if (outFile != stdout) fclose(outFile);
	return 0;
}
